<?LassoScript
	
	//
	// Caching Tags
	//
	// Copyright 2008, LassoSoft, LLC
	//
	// Installation - Place in the LassoStartup folder in the Lasso
	// Professional 8 application folder.  Suports Lasso 8.5 or higher.
	//
	// These cache tags are a drop-in replacement for the cache tags in
	// Lasso 8.5.5 and earlier. These tags are faster and more stable on
	// high traffic servers. Multiple global variables are used for
	// cache storage rather than a single global map. Atomic sections
	// protect write operations forcing them into serial order.
	// 
	// Thanks to Johan Slve for contributions to the cache tags and for
	// earlier implementations of garbage collecion and thread safety.
	//
	// Note - Cached objects are not serialized so the -NoSerialize
	// option is no longer recognized. The ability to encrypt cached
	// objects is not present and the -Key option is no longer
	// recognized. This ability was intended for use on shared servers,
	// but is rendered largely moot by the multi-site capabilities in
	// Lasso 8.
	//
	// [Cache] ... [/Cache] - Requires a -Name which names the
	// particular cache. The tag can be passed an -Expires parameter
	// which specifies the number of seconds the cache should persist
	// for. If the current cached content is older than this number of
	// seconds then the new content is stored. The [Cache] ... [/Cache]
	// container will only execute its contents if necessary. Returns
	// the current cached value.
	//
	// [Cache: -Name='name', -Expires='seconds')]
	//   ... String Contents ...
	// [/Cache]
	//	
	// [Cache_Object] - Requires the same paramters as the container
	// tag. This tag will ->Invoke its -Content parameter if necessary.
	// Returns the current cached value.
	
	// [Cache_Object: -Name='name', -Expires='seconds', -Content='value']
	//	
	// An optional -Condition parameter can be specified as an
	// alternative to the -Expires parameter in either tag. If
	// -Condition=True is specified then the cache will be refreshed
	// with new content. If -Condition=False is specified then the
	// current value in the cache will be returned.  An optional -Log
	// parameter turns on logging.
	//
	// By default the tags work with global storage. An optional
	// -Session parameter names a session in which the cache should be
	// stored. This allows content to be cached per user. If
	// -UseGlobal=False is specified then page variables will be used
	// for cache storage.
	//
	// The remainder of the tags perform explicit cache operations
	// without any expiration or condition. [Cache_Store] stores the
	// specified content in the cache and returns no value.  The value
	// is invoked before it is stored. [Cache_Fetch] fetches the current
	// value for the cache. [Cache_Empty] empties the current value from
	// the cache forcing a refresh the next time it is loaded, but
	// maintains the current fetch/store statistics. [Cache_Delete]
	// deletes the cache altogether including its statistics.
	// [Cache_Exists] reports true if a cache has a value.
	//
	// [Cache_Store: -Name='name', -Content='value']
	// [Cache_Fetch: -Name='name']
	// [Cache_Empty: -Name='name']
	// [Cache_Delete: -Name='name']
	// [Cache_Exists: -Name='name']
	//
	// These tags accpets the same -Session and -UseGlobal parameters as
	// the [Cache] ... [/Cache] container and [Cache_Object] tag.  The
	// [Cache_Exists] tag can be passed an -Expires parameter or
	// -Condition parameter in order to simulate the check performed by
	// the [Cache] ... [/Cache] container or [Cache_Object] tag.
	//
	// [Cache_Maintenance] - The garbage collection routines run in an
	// asynchronous thread. The garbage collector adds up the space
	// required for each cached item by serializing it. If the size
	// exceeds the specified limit then items are emptied in order of
	// their efficiency until the remaining items occupy less than 2/3
	// of the specified limit. The maintenance routine will be run nightly
	// in order to target a low traffic time of the day.
	//
	// [Cache_Internal] - A subroutine which should not be called directly.
	//
	
	// Note - [Cache] ... [/Cache] container defined in LassoStartup
	
	define_tag('object', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='content',
			-optional='expires',
			-optional='condition',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;

		local('c_exists') = (\cache_exists->run(-params=params));
		local('c_store') = true;
		local('c_content') = '';
		
		if(local_defined('content') && ((local('condition') == true) || (#c_exists == false)));
			if(#c_useglobal);
				local('count') = integer(globals->find(#c_prefix + 'store'));
				thread_atomic(globals->find(#c_prefix + 'content'));
					#c_store = (#count == integer(globals->find(#c_prefix + 'store')));
					if(#c_store);
						#c_content = #content->invoke;
						globals->insert((#c_prefix + 'content') = #c_content);
						globals->insert((#c_prefix + 'date') = date->format('%Q %T'));
						globals->insert((#c_prefix + 'store') = integer(globals->find(#c_prefix + 'store')) + 1);
						#c_log ? log_warning('Cache: Storing ' + #c_type + ' Object "' + #c_name + '".');
						return(#c_content);
					else;
						#c_log ? log_warning('Cache: Waiting for ' + #c_type + ' Object "' + #c_name + '".');
					/if;
				/thread_atomic;
			else;
				#c_content = string(run_children);
				vars->insert((#c_prefix + 'content') = #c_content);
				vars->insert((#c_prefix + 'date') = date->format('%Q %T'));
				vars->insert((#c_prefix + 'store') = integer(vars->find(#c_prefix + 'store')) + 1);
				if(#c_session);
					session_addvar(-name=#session, #c_prefix + 'content', #c_prefix + 'date', #c_prefix + 'fetch', #c_prefix + 'store');
				/if;
				#c_log ? log_warning('Cache: Storing ' + #c_type + ' Object "' + #c_name + '".');
				return(#c_content);
			/if;
		else;
			#c_log ? log_warning('Cache: Fetching ' + #c_type + ' Object "' + #c_name + '".');
		/if;

		// Return the content
		if(#c_useglobal);
			thread_atomic(globals->find(#c_prefix + 'fetch'));
				globals->insert((#c_prefix + 'fetch') = integer(globals->find(#c_prefix + 'fetch')) + 1);
			/thread_atomic;
			return(globals->find(#c_prefix + 'content'));
		else;
			vars->insert((#c_prefix + 'fetch') = integer(vars->find(#c_prefix + 'fetch')) + 1);
			return(vars->find(#c_prefix + 'content'));
		/if;

	/define_tag;
	
	define_tag('store', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-required='content', -copy,
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;
				
		#c_log ? log_warning('Cache: Storing ' + #c_type + ' Cache "' + #c_name + '".');

		// Generate content, date, fetch, store variables
		if(#c_useglobal);
			thread_atomic(globals->find(#c_prefix + 'content'));
				globals->insert((#c_prefix + 'content') = #content->invoke);
				globals->insert((#c_prefix + 'date') = date->format('%Q %T'));
				globals->insert((#c_prefix + 'store') = integer(globals->find(#c_prefix + 'store')) + 1);
			/thread_atomic;
		else;
			vars->insert((#c_prefix + 'content') = #content->invoke);
			vars->insert((#c_prefix + 'date') = date->format('%Q %T'));
			vars->insert((#c_prefix + 'store') = integer(vars->find(#c_prefix + 'store')) + 1);
			if(#c_session);
				session_addvar(-name=#session, #c_prefix + 'content', #c_prefix + 'date', #c_prefix + 'fetch', #c_prefix + 'store');
			/if;
		/if;			

	/define_tag;

	define_tag('fetch', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;
		
		#c_log ? log_warning('Cache: Fetching ' + #c_type + ' Cache "' + #c_name + '".');

		// Return the content
		if(#c_useglobal);
			thread_atomic(globals->find(#c_prefix + 'fetch'));
				globals->insert((#c_prefix + 'fetch') = integer(globals->find(#c_prefix + 'fetch')) + 1);
				return(globals->find(#c_prefix + 'content'));
			/thread_atomic;
		else;
			vars->insert((#c_prefix + 'fetch') = integer(vars->find(#c_prefix + 'fetch')) + 1);
			return(vars->find(#c_prefix + 'content'));
		/if;

	/define_tag;

	define_tag('empty', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;
		
		#c_log ? log_warning('Cache: Emptying ' + #c_type + ' Cache "' + #c_name + '".');

		// Empty the content, delete the date
		if(#c_useglobal);
			thread_atomic(globals->find(#c_prefix + 'content'));
				globals->insert((#c_prefix + 'content') = null) & remove(#c_prefix + 'date');
			/thread_atomic;
		else;
			vars->remove(#c_prefix + 'content') & remove(#c_prefix + 'date');
			if(#c_session);
				session_removevar(-name=#session, #c_prefix + 'content', #c_prefix + 'date');
			/if;
		/if;

	/define_tag;

	define_tag('delete', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;
		
		#c_log ? log_warning('Cache: Deleting ' + #c_type + ' Cache "' + #c_name + '".');

		// Remove globals, vars, and session vars
		if(#c_useglobal);
			thread_atomic(globals->find(#c_prefix + 'content'));
				globals->remove(#c_prefix + 'content') & remove(#c_prefix + 'date') & remove(#c_prefix + 'store');
			/thread_atomic;
			thread_atomic(globals->find(#c_prefix + 'fetch'));
				globals->remove(#c_prefix + 'fetch');
			/thread_atomic;
		else;
			vars->remove(#c_prefix + 'content') & remove(#c_prefix + 'date') & remove(#c_prefix + 'fetch') & remove(#c_prefix + 'store');
			if(#c_session);
				session_removevar(-name=#session, #c_prefix + 'content', #c_prefix + 'date', #c_prefix + 'fetch', #c_prefix + 'store');
			/if;
		/if;	

	/define_tag;

	define_tag('exists', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='expires',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		\cache_internal->eval;
				
		// False if the cache has no content
		if(#c_useglobal);
			globals->find(#c_prefix + 'content') === null ? return(false);
		else;
			vars->find(#c_prefix + 'content') === null ? return(false);
		/if;
		
		// True if no expiration is specified.
		!local_defined('expires') || (integer(#expires) <= 0) ? return(true);
		
		// Fetch the date of the last store
		if(#c_useglobal);
			local('date') = globals->find(#c_prefix + 'date');
		else;
			local('date') = vars->find(#c_prefix + 'date');
		/if;
	
		// False if the date is not defined
		(#date !>> '-' || #date !>> ':') ? return(false);
		// Check the expiration

		return((date(#date)->add(-second=integer(#expires)) &) > date);
		
	/define_tag;

	define_tag('internal', -priority='replace', -namespace='_global_cache_',
			-required='name',
			-optional='expires',
			-optional='session',
			-optional='useglobal',
			-optional='log');

		// Load Preferences
		if((globals->keys !>> '_Cache_Preference') || (globals->find('_Cache_Preference')->type != 'map'));
			thread_atomic(globals->find('_Cache_Preference'));
				globals->insert('_Cache_Preference' = Cache_Preferences);
			/thread_atomic;
		/if;
		
		// Bail if caches are disabled
		global('_Cache_Preference')->(find: '_cache_enable') == 'N' ? return(local('content'));

		// Parameters
		local('c_useglobal') = local_defined('useglobal') && (#useglobal === true);
		local('c_session') = local_defined('session') && (#c_useglobal == false);
		local('c_useglobal') = #c_session == false;
		local('c_log') = local_defined('log') && (#log !== false);

		if(#c_session);
			local('c_key') = session_id(-name=#session);
			(#c_key == '') ? local('c_key') = 'session';
			local('c_type') = 'Session';
		else(#c_useglobal == false);
			local('c_key') = 'local';
			local('c_type') = 'Local';
		else(global('_Cache_Preference')->(find: '_cache_host') == 'Y');
			local('c_key') = server_name;
			local('c_type') = 'Global';
		else;
			local('c_key') = 'global';
			local('c_type') = 'Global';
		/if;
		
		// Append c_key to cache name for variable c_prefix
		local('c_prefix') = '_cache_' + #name + '_' + #c_key + '_';
		local('c_name') = #name;
		
	/define_tag;

	define_tag('maintenance', -priority='replace', -async, -namespace='_global_cache_', -optional='log');
		globals >> '_cache_maintenance_' ? return;
		globals->insert('_cache_maintenance_');

		if((globals->keys !>> '_Cache_Preference') || (globals->find('_Cache_Preference')->type != 'map'));
			thread_atomic(globals->find('_Cache_Preference'));
				globals->insert('_Cache_Preference' = Cache_Preferences);
			/thread_atomic;
		/if;
		
		// Quit if no limit is set
		local('limit') = integer(globals->find('_Cache_Preference')->find('_cache_totalstoragelimit'));
		#limit <= 0 ? return;
		local('threshold') = #limit * 0.67; // Threshold is 2/3 of the limit

		// Calculate fetch/fetch+store percentages for all caches
		local('stats') = array;
		local('caches') = array;
		local('stores') = map;
		local('fetches') = map;
		iterate(globals->keys, local('key'));
			!#key->beginswith('_cache_') ? loop_continue;
			if(#key->endswith('_content'));
				#key->removeleading('_cache_')&removetrailing('_content');
				#caches->insert(#key);
			else(#key->endswith('_store'));
				local('value') = decimal(globals->find(#key));
				#key->removeleading('_cache_')&removetrailing('_store');
				#stores->insert(#key = #value);
			else(#key->endswith('_fetch'));
				local('value') = decimal(globals->find(#key));
				#key->removeleading('_cache_')&removetrailing('_fetch');
				#fetches->insert(#key = #value);
			/if;
		/iterate;
		
		iterate(#caches, local('key'));
			if((#fetches !>> #key) || #fetches->find(#key) == 0 || (#stores !>> #key) || #stores->find(#key) == 0 || (#fetches->find(#key) < #stores->find(#key)));
				#stats->insert(0 = #key);
			else;
				#stats->insert((#fetches->find(#key) - #stores->find(#key)) / (#fetches->find(#key)) = #key);
			/if;
		/iterate;
		#stats->sort(false); // descending
		
		// Calculate effective sizes
		local('deletes') = array;
		local('total') = 0;
		iterate(#stats, local('temp'));
			local('key') = #temp->second;
			if(#total > #limit);
				// Already over the limit, delete
				#deletes->insert(#key);
			else;
				// Still under the limit, check sizes
				thread_atomic(globals->find('_cache_' + #key + '_content'));
					local('type') = globals->find('_cache_' + #key + '_content')->type;
					if((#type == 'string') || (#type == 'bytes'));
						local('size') = globals->find('_cache_' + #key + '_content')->size;
					else;
						local('size') = serialize(globals->find('_cache_' + #key + '_content'))->size;
					/if;
				/thread_atomic;
				#total += #size;
				if(#total > #threshold);
					#deletes->insert(#key);
				/if;
			/if;
		/iterate;
		
		// Process deletes if we are over limit
		if(#total > #limit);
			log_warning('Cache: Garbage Collection Deleted ' + #deletes->size + ' Most Inefficient Caches.');
			iterate(#deletes, local('key'));
				thread_atomic(globals->find('_cache_' + #key + '_content'));
					globals->remove('_cache_' + #key + '_content') & remove('_cache_' + #key + '_date') & remove('_cache_' + #key + '_store');
				/thread_atomic;
				thread_atomic(globals->find('_cache_' + #key + '_fetch'));
					globals->remove('_cache_' + #key + '_fetch');
				/thread_atomic;
			/iterate;
		/if;
		
		globals->remove('_cache_maintenance_');
	/define_tag;

?>
