<?LassoScript

	/*
		Library of routines to perform miscellaneous table/database functions on internal mySQL datasource
		Implements:
			[Database_CreateTable]
			[Database_RemoveTable]
			[Database_RemoveField]
				[Database_RemoveColumn]
			[Database_ChangeField]
				[Database_ChangeColumn]
			[Database_CreateField]
				[Database_CreateColumn]
			[Database_TableColumn] data type

		Problems:
			Not all functions are smart enough to take params in unusual order yet
			Does not remove/error detect spaces, decimals in table/column name inputs
			Docs need cleanup

		History:
			06/15/2004	AJR - Moved to pure LassoScript
			          	AJR - Renamed Table_Column to Database_TableColumn
			          	AJR - Moved to on-demand loader (this file)
			10/22/2003	AJR - removed dependencies on named_param.
			06/14/2001	ABK - added fully-qualified table names and quoted entities
			05/31/2001	GLH - renamed tags to match Fletcher's new spec
			          	GLH - changed semantics of tags to match Fletcher's new spec
			          	GLH - Database_CreateTable always generates a default 'id' column
			          	GLH - added synonym [Database_RemoveColumn]
			03/14/2001	GLH - final code, still need to execute SQL
			          	GLH - finished code, added [Table_Delete], SQL execution tested
			03/12/2001	GLH - more coding
			03/05/2001	GLH - initial coding
	*/

	/*
		Database_RemoveTable:	custom tag to delete a mySQL table

			[Database_RemoveTable:
				-Database='site',    // required: name of database from which to remove table
				-Table='table_name'  // required: name of table to delete
			]

		Problems:

		Inputs:
			-Database	name of database from which to delete table
			-Table		name of table to delete

		Output:
			No text output, but side-effect of destroying a table
	*/

	Define_Tag: 'RemoveTable', -namespace='_global_Database_', -priority='replace';
		// OK, here we go: execute the SQL statement
		Inline:
			-Database=(local: 'Database'),
			-sql='DROP TABLE IF EXISTS `' + (local: 'Database') + '`.`' + (local: 'Table') + '`';
		/Inline;
	/Define_Tag;



	/*
		Database_CreateTable:	custom tag to create a mySQL table from scratch

			[Database_CreateTable:
				-Database='site',             // required, name of database in which to create new table
				-Table='table_name',          // required: name of table to make
				(Database_TableColumn: ...),  // required: column to create in table
				(Database_TableColumn: ...),  // required: column to create in table
				optional -Temporary           // optional: make table temporary
			]

		Problems:

		Inputs:
			-Database	required	name of database in which table will be created
			-Table		required	name of table to create
			-Temporary	optional	flag for making this a temp table

		Output:
			No text output, but side-effect of creating a table with a single auto_increment field 'id'
	*/

	Define_Tag: 'CreateTable', -namespace='_global_Database_', -priority='replace';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		// Grab the required table name
		Local: 'tableName'=(local: 'Table');

		// set default value for temporary table
		Local: 'isTemp'=false;

		Loop: params->size;
			If: params->get: loop_count == '-Temporary';
				Local: 'isTemp'=true;
			/If;
		/Loop;

		// Output SQL statement to create the table
		Local: 'sqlText' = 'CREATE ';
		If: #isTemp;
			Local: 'sqlText' = #sqlText + 'TEMPORARY ';
		/If;
		Local: 'sqlText' = #sqlText + 'TABLE `' + (local: 'Database') + '`.`' + #tableName + '` (id INT PRIMARY KEY AUTO_INCREMENT)';

		// OK, here we go: execute the SQL statement
		Inline:
			-Database=(local: 'Database'),
			-sql=#sqlText;
		/Inline;
	/Define_Tag;



	/*
		Database_TableColumn: custom data type to assist with building columns

			[Database_TableColumn:
				'column_name',    // required: name of field
				'column_type',    // required: SQL type: INTEGER, FLOAT, TEXT (http://dev.mysql.com/doc/mysql/en/CREATE_TABLE.html)
				'default_value',  // optional: default value for this field
				-AutoIncrement,   // optional: if field is to autoincrement on each new Add
				-Key,             // optional: if this is a key field
				-Null,            // optional: if NULL is allowed in this field
				-NotNull          // optional: if NULL is denied for this field
			]

		Problems:

		Inputs:

		Output:
			[OutputSQL]	member function outputs SQL text sufficient for defining this column
	*/

	Define_Type: 'TableColumn', -namespace='_global_Database_';
		Local: 'colIncrement' = false;
		Local: 'colKey' = false;
		Local: 'colNull' = false;
		Local: 'colNotNull' = false;

		Local: 'colName'=null;
		Local: 'colType'=null;
		Local: 'colDefault'=null;

		Local: 'colName' = params->get: 1;
		If: params->size > 1;
			Local: 'colType' = params->get: 2;
		/If;
		If: params->size > 2 && (params->(get: 3) != null);
			If: ((params->get: 3)->get: 1) != '-';
				Local: 'colDefault' = params->get: 3;
			/If;
		/If;
		Loop: params->size;
			If: params->get: loop_count == '-AutoIncrement';
				Local: 'colIncrement' = true;
			Else: params->get: loop_count == '-Key';
				Local: 'colKey' = true;
			/If;
			If: params->get: loop_count == '-Null';
				Local: 'colNull' = true;
			/If;
			If: params->get: loop_count == '-NotNull';
				Local: 'colNotNull' = true;
			/If;
		/Loop;

		Define_Tag: 'SetKey';
			self->'colKey'=params->(get: 1);
		/Define_Tag;

		Define_Tag: 'SetNull';
			self->'colNull'=params->(get: 1);
		/Define_Tag;

		Define_Tag: 'SetNotNull';
			self->'colNotNull'=params->(get: 1);
		/Define_Tag;

		Define_Tag: 'SetAutoIncrement';
			self->'colIncrement'=params->(get: 1);
		/Define_Tag;

		Define_Tag: 'OutputSQL';
			Local: 'sqlText' = '`' + (self->'colName') + '` ' + (String_Uppercase: self->'colType');
			If: self->'colNull';
				Local: 'sqlText' = #sqlText + ' NULL';
			/If;
			If: self->'colNotNull';
				Local: 'sqlText' = #sqlText + ' NOT NULL';
			/If;
			If: (self->'colDefault' != '-Null') && (self->'colDefault' != '');
				Local: 'sqlText' = #sqlText + ' DEFAULT \'' + self->'colDefault' + '\'';
			/If;
			If: self->'colKey';
				Local: 'sqlText' = #sqlText + ' PRIMARY KEY';
			/If;
			If: self->'colIncrement';
				Local: 'sqlText' = #sqlText + ' AUTO_INCREMENT';
			/If;
			Return: #sqlText;
		/Define_Tag;
	/Define_Type;



	/*
		Database_RemoveField:	custom tag to delete a column from mySQL table

			[Database_RemoveField:
				-Database='site',     // required: name of database in which the table resides
				-Table='table_name',  // required: name of table to make
				-Field='column_name'  // required: name of column (field) to delete from table schema
			]

		Problems:

		Inputs:
			-Database	required	name of database in which the table resides
			-Table		required	name of table to make
			-Field		required	name of column (field) to delete from table schema

		Output:
			No text output, but side-effect of deleting a column
	*/

	Define_Tag: 'RemoveField', -namespace='_global_Database_', -priority='replace';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		PreCondition: (local: 'Field') != '';
		// Grab the required table name
		Local: 'tableName'=(local: 'Table');

		// Grab the required column name
		Local: 'colName'=(local: 'Field');

		// Execute SQL statement to delete the column
		Inline:
			-Database=(local: 'Database'),
			-sql='ALTER TABLE `' + (local: 'Database') + '`.`' + #tableName + '` DROP COLUMN `' + #colName + '`';
		/Inline;
		// Forces MySQL to see new field
		Inline:
			-database=(local: 'Database'),
			-sql='FLUSH TABLE `' + (local: 'Database') + '`.`' + #tableName + '`';
		/inline;
	/Define_Tag;



	/*
		Database_RemoveColumn:	synonym of [Database_RemoveField]

			[Database_RemoveField:
				-Database='site',      // required: name of database in which the table resides
				-Table='table_name',   // required: name of table to make
				-Column='column_name'  // required: name of column (field) to delete from table schema
			]

		Problems:

		Inputs:
			-Database	required	name of database in which the table resides
			-Table		required	name of table to make
			-Column		required	name of column (field) to delete from table schema

		Output:
			No text output, but side-effect of deleting a column
	*/

	Define_Tag: 'RemoveColumn', -namespace='_global_Database_', -priority='replace';
		Database_RemoveField:
			-Database=(local: 'Database'),
			-Table=(local: 'Table'),
			-Field=(local: 'Column');
	/Define_Tag;



	/*
		Database_ChangeField:	custom tag to modify a column in mySQL table

			[Database_ChangeField:
				-Database='site',         // required: name of database in which table resides
				-Table='table_name',      // required: name of table to modify
				-Original='column_name',  // required: name of column (field) to modify
				-Field='firstname',       // requiredL name of new fieldname
				-Type='VARCHAR(25)',
				-Default='NoNameGiven',
				-AutoIncrement,
				-Key,
				-Null,
				-NotNull
			]

		Problems:

		Inputs:

		Output:
	*/

	Define_Tag: 'ChangeField', -namespace='_global_Database_', -priority='replace';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		PreCondition: (local: 'Original') != '';
		PreCondition: (local: 'Field') != '';
		PreCondition: (local: 'Type') != '';

		// Grab the column definition
		/*
		Loop: params->size;
			If: (params->get: loop_count)->type == 'Database_TableColumn';
				Local: 'theCol'=params->get: loop_count;
			/If;
		/Loop;
		*/

		Local: 'theCol'=(Database_TableColumn:
			(local: 'Field'),
			(local: 'Type'),
			(local: 'Default'));

		// Determine the params which have no value
		Loop: params->size;
			If: params->get: loop_count == '-AutoIncrement';
				#theCol->SetAutoIncrement: true;
			/If;
			If: params->get: loop_count == '-Key';
				#theCol->SetKey: true;
			/If;
			If: params->get: loop_count == '-Null';
				#theCol->SetNull: true;
			/If;
			If: params->get: loop_count == '-NotNull';
				#theCol->SetNotNull: true;
			/If;
		/Loop;

		// Execute SQL statement to delete the column
		Inline:
			-Database=(local: 'Database'),
			-sql='ALTER TABLE `' + (local: 'Database') + '`.`' + (local: 'Table') + '` CHANGE COLUMN `' + (local: 'Original') + '` ' + #theCol->OutputSQL;
		/inline;
		// Forces MySQL to see new field
		Inline:
			-database=(local: 'Database'),
			-sql='FLUSH TABLE `' + (local: 'Database') + '`.`' + (local: 'Table') + '`';
		/inline;
	/Define_Tag;



	/*
		Database_ChangeColumn:	custom tag to modify a column in mySQL table

			[Database_ChangeColumn:
				-Database='site',         // required: name of database in which table resides
				-Table='table_name',      // required: name of table to modify
				-Original='column_name',  // required: name of column (field) to modify
				-Column='firstname',      // requiredL name of new fieldname
				-Type='VARCHAR(25)',
				-Default='NoNameGiven',
				-AutoIncrement,
				-Key,
				-Null,
				-NotNull
			]

		Problems:

		Inputs:

		Output:
	*/

	Define_Tag: 'ChangeColumn', -namespace='_global_Database_';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		PreCondition: (local: 'Original') != '';
		PreCondition: (local: 'Column') != '';
		PreCondition: (local: 'Type') != '';

		// Grab the column definition
		/*
		Loop: params->size;
			If: (params->get: loop_count)->type == 'Database_TableColumn';
				Local: 'theCol'=params->get: loop_count;
			/If;
		/Loop;
		*/

		Local: 'theCol'=(Database_TableColumn:
			(local: 'Column'),
			(local: 'Type'),
			(local: 'Default'));

		// Determine the params which have no value
		Loop: params->size;
			If: params->get: loop_count == '-AutoIncrement';
				#theCol->SetAutoIncrement: true;
			/If;
			If: params->get: loop_count == '-Key';
				#theCol->SetKey: true;
			/If;
			If: params->get: loop_count == '-Null';
				#theCol->SetNull: true;
			/If;
			If: params->get: loop_count == '-NotNull';
				#theCol->SetNotNull: true;
			/If;
		/Loop;

		// Execute SQL statement to delete the column
		Inline:
			-Database=(local: 'Database'),
			-sql='ALTER TABLE `' + (local: 'Database') + '`.`' + (local: 'Table') + '` CHANGE COLUMN `' + (local: 'Original') + '` ' + #theCol->OutputSQL;
		/inline;
		// Forces MySQL to see new field
		Inline:
			-database=(local: 'Database'),
			-sql='FLUSH TABLE `' + (local: 'Database') + '`.`' + (local: 'Table') + '`';
		/inline;
	/Define_Tag;



	/*
		Database_CreateField:	custom tag to insert new column into mySQL table

			[Database_CreateField:
				-Database='site',        // required: name of database which holds the table in question
				-Table='table_name',     // required: name of table to modify
				-Field='fieldname',      // required: name of field to add
				-Type='VARCHAR(25)',
				-Default='NoNameGiven',
				-AutoIncrement,
				-Key,
				-Null,
				-NotNull,
				-BeforeFirst,            // optional: if column should be inserted at front
				-AfterField='fieldname'  // optional: name of column to insert after
			]

		Problems:
			If user specifies both -BeforeFirst and -AfterField, both are written to SQL statement

		Inputs:

		Output:
			No text output, but side-effect of adding a column
	*/

	Define_Tag: 'CreateField', -namespace='_global_Database_';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		PreCondition: (local: 'Field') != '';

		// Grab the required table name
		Local: 'tableName'=(local: 'Table');

		Local: 'after'=(local: 'AfterField');
		Local: 'beforeFirst' = false;

		Local: 'theCol'=(Database_TableColumn:
			(local: 'Field'),
			(local: 'Type'),
			(local: 'Default'));

		// Determine the params which have no value
		Loop: params->size;
			If: params->get: loop_count == '-AutoIncrement';
				#theCol->SetAutoIncrement: true;
			/If;
			If: params->get: loop_count == '-Key';
				#theCol->SetKey: true;
			/If;
			If: params->get: loop_count == '-Null';
				#theCol->SetNull: true;
			/If;
			If: params->(get: loop_count) == '-NotNull';
				#theCol->SetNotNull: true;
			/If;
			If: params->(get: loop_count) == '-BeforeFirst';
				Local: 'beforefirst' = true;
			/If;
		/Loop;

		Local: 'sqlText' = 'ALTER TABLE `' + (local: 'Database') + '`.`' + #tableName + '` ADD COLUMN ' + #theCol->OutputSQL;
		If: #after != '';
			Local: 'sqlText' = #sqlText + ' AFTER `' + #after + '`';
		/If;
		If: #beforeFirst;
			Local: 'sqlText' = #sqlText + ' FIRST';
		/If;

		// Now we execute the SQL code to add the column
		Inline:
			-Database=(local: 'Database'),
			-sql=#sqlText;
		/Inline;
		// Forces MySQL to see new field
		Inline:
			-database=(local: 'Database'),
			-sql='FLUSH TABLE `' + (local: 'Database') + '`.`' + #tableName + '`';
		/inline;
	/Define_Tag;



	/*
		Database_CreateColumn:	custom tag to insert new column into mySQL table

			[Database_CreateColumn:
				-Database='site',        // required: name of database which holds the table in question
				-Table='table_name',     // required: name of table to modify
				-Column='fieldname',     // required: name of field to add
				-Type='VARCHAR(25)',
				-Default='NoNameGiven',
				-AutoIncrement,
				-Key,
				-Null,
				-NotNull,
				-BeforeFirst,            // optional: if column should be inserted at front
				-AfterColumn='fieldname' // optional: name of column to insert after
			]

		Problems:
			If user specifies both -BeforeFirst and -AfterColumn, both are written to SQL statement

		Inputs:

		Output:
			No text output, but side-effect of adding a column
	*/

	Define_Tag: 'CreateColumn', -namespace='_global_Database_';
		PreCondition: (local: 'Database') != '';
		PreCondition: (local: 'Table') != '';
		PreCondition: (local: 'Column') != '';

		// Grab the required table name
		Local: 'tableName'=(local: 'Table');

		Local: 'after'=(local: 'AfterColumn');
		Local: 'beforeFirst' = false;

		Local: 'theCol'=(Database_TableColumn:
			(local: 'Column'),
			(local: 'Type'),
			(local: 'Default'));

		// Determine the params which have no value
		Loop: params->size;
			If: params->get: loop_count == '-AutoIncrement';
				#theCol->SetAutoIncrement: true;
			/If;
			If: params->get: loop_count == '-Key';
				#theCol->SetKey: true;
			/If;
			If: params->get: loop_count == '-Null';
				#theCol->SetNull: true;
			/If;
			If: params->(get: loop_count) == '-NotNull';
				#theCol->SetNotNull: true;
			/If;
		/Loop;

		Local: 'sqlText' = 'ALTER TABLE `' + (local: 'Database') + '`.`' + #tableName + '` ADD COLUMN ' + #theCol->OutputSQL;
		If: #after != '';
			Local: 'sqlText' = #sqlText + ' AFTER `' + #after + '`';
		/If;
		If: #beforeFirst;
			Local: 'sqlText' = #sqlText + ' FIRST';
		/If;

		// Now we execute the SQL code to add the column
		Inline:
			-Database=(local: 'Database'),
			-sql=#sqlText;
		/Inline;
		// Forces MySQL to see new field
		Inline:
			-database=(local: 'Database'),
			-sql='FLUSH TABLE `' + (local: 'Database') + '`.`' + #tableName + '`';
		/inline;
	/Define_Tag;

?>