<?LassoScript

	// Valid Tags
	//
	// Note - These are convenience tags which check for the most common formatting for each value.
	// According to the RFCs and other standards a great many values which these tags return false
	// for will be technically valid.  The credit card tag checks the ROT-13 algorithm which is a
	// good first check for whether a credit card is valid.
	//
	// [Valid_Email]
	//
	// [Valid_URL]
	//
	// [Valid_Date]
	//
	// [Valid_CreditCard]
	//

	Define_Tag: 'Email', -Priority='Replace',
			-Namespace='_global_Valid_',
			-Required='email',
			-Optional='standarddomains',
			-Optional='domain',
			-Optional='hostname';

		error_push;
		
		local: '_output' = null;

		protect;

			local: '_email' = (string_findregexp:
				(string: #email)->trim &,
				-find='[^<>@ ]+@[A-Za-z0-9\\-_]+\\.[A-Za-z0-9\\-_\\.]+');

			if: (#_output === null) && (#_email->type == 'array') && (#_email->size == 1) && (#_email->(get: 1) == #email);
				local: '_hostname' =  #_email->(get: 1)->(split: '@')->(get: 2);
				local: '_domain' =  #_hostname->(split: '.')->last;
			else;
				local: '_output' = false;
			/if;

			if: (#_output === null) && (local_defined: 'standarddomains') && ((array: 'com','gov','mil','net','org','int') !>> #_domain);
				local: '_output' = false;
			/if;

			if: (#_output === null) && (local_defined: 'domain');
				if: (#domain->(isa: 'array')) && (#domain !>> #_domain);
					local: '_output' = false;
				else: ! #domain->(isa: 'array');
					local: 'temp' = (string: #domain)->(split: ',');
					if: (#temp->(isa: 'array')) && (#temp !>> #_domain);
						local: '_output' = false;
					/if;
				/if;
			/if;

			if: (#_output === null) && (local_defined: 'hostname') && (#hostname != #_hostname);
				local: '_output' = false;
			/if;

			if: (#_output === null);
				local: '_output' = true;
			/if;
			
			handle_error;
				log_warning: 'Valid_Email Error: ' (error_code) ': ' (error_msg);
				local: '_output' = false;
			/handle_error;

		/protect;

		error_pop;
	
		return: @#_output;
			
	/Define_Tag;



	Define_Tag: 'URL', -Priority='Replace',
			-Namespace='_global_Valid_',
			-required='url',
			-optional='protocol';

		local: '_url' = (string: #url);

		#_url->trim;

		Local: 'parse' = (string_findregexp:
			#_url,
			-find='([^:]+):(/?/?)([:@A-Za-z0-9-_\\.]+)(/?[^ \\t]*)');

		if: (#parse->size < 5);
			return: false;
		/if;

		if: (#_url != (#parse->(get:2) + ':' + #parse->(get:3) + #parse->(get:4) + #parse->(get:5)));
			return: false;
		/if;

		if: (local_defined: 'protocol');
			if: (#protocol->(isa: 'array')) && (#protocol !>> #parse->(get:2));
				return: false;
			else: ! #protocol->(isa: 'array');
				local: 'temp' = #protocol->(split: ',');
				if: (#temp->(isa: 'array')) && (#temp !>> #parse->(get:2));
					return: false;
				/if;
			/if;
		/if;

		return: true;

	/Define_Tag;



	Define_Tag: 'Date', -Priority='Replace',
			-Namespace='_global_Valid_',
			-Required='date',
			-Optional='format';

		// Empty input
		#date == null ? return: false;
		(string: #date)->trim & == '' ? return: false;

		// Parse date
		if: (local_defined: 'format');
			local: 'parse' = (date: #date, -format=#format);
		else;
			local: 'parse' = (date: #date);
		/if;
		
		// Invalid dates
		(#parse->type == 'null') ? return: false;
		((string: #parse)->size == 0) ? return: false;
		(#parse >> '0000') ? return: false;
		
		// Strict check of format against original date (allowing for leading zeroes or spaces)
		if: (local_defined: 'format');
			(string: #date) == #parse->(format: (string_replaceregexp: #format, -find='%-?_?', -replace='%')) ? return: true;
			(string: #date) == #parse->(format: (string_replaceregexp: #format, -find='%-?_?', -replace='%_')) ? return: true;
			(string: #date) == #parse->(format: (string_replaceregexp: #format, -find='%-?_?', -replace='%-')) ? return: true;
			return: false;
		/if;
		
		return: true;

	/Define_Tag;



	Define_Tag: 'CreditCard', -Priority='Replace',
			-Namespace='_global_Valid_',
			-Required='card',
			-optional='type';

		//
		// Clean the card number and check its length
		//

		local: '_card' = (string_replaceregexp: #card, -find='\\D', -replace='');

		if: (#_card->size < 13);
			return: false;
		/if;

		//
		// Check the ROT-13 checksum for the card number
		//

		local: 'rev' = #_card;
		#rev->reverse;
		local: 'checksum' = 0;
		iterate: #rev, (local: 'digit');
			if: (loop_count % 2) == 0;
					#checksum += (2 * (integer: #digit));
				if: (integer: #digit) >= 5;
					#checksum -= 9;
				/if;
			else;
				#checksum += (integer: #digit);
			/if;
		/iterate;

		if: (#checksum % 10) != 0;
			return: false;
		/if;

		//
		// If a type is specified check the first character and length of the credit card
		//

		if: (local_defined: 'type');
			if: (#type >> 'visa');
				if: ((#_card->size != 13) && (#_card->size != 16)) || (#_card->(get:1) != '4');
					return: false;
				/if;
			else: (#type >> 'mastercard');
				if: (#_card->size != 16) || (#_card->(get:1) != '5');
					return: false;
				/if;
			else: (#type >> 'discover');
				if: (#_card->size != 16) || (#_card->(get:1) != '6');
					return: false;
				/if;
			else: (#type >> 'amex');
				if: (#_card->size != 15) || (#_card->(get:1) != '3');
					return: false;
				/if;
			else: (#type >> 'diners');
				if: (#_card->size != 14) || (#_card->(get:1) != '3');
					return: false;
				/if;
			else: (#type >> 'jcb');
				if: (#_card->size != 15) && (#_card->size != 16);
					return: false;
				/if;
			/if;
		/if;

		return: true;

	/Define_Tag;

?>