<?lassoscript

	/*
		Library of routines to perform miscellaneous functions that used to be performed in Java modules
		Implements:
			[WAP_IsEnabled]	true if the visiting browser is a WAP-enabled (HDML) device, false if not
			[WAP_MaxColumns]	integer indicates how many horizontal characters the device can display
			[WAP_MaxRows]	integer indicates how many vertical characters the device can display
			[WAP_MaxHorzPixels]	integer indicates how many horizontal pixels the device can display
			[WAP_MaxVertPixels]	integer indicates how many vertical pixels the device can display
			[WAP_MaxButtons]	integer indicates how many soft-buttons the device has

		Problems:
			[Include_CGI] rewritten to use CURL tags, needs testing.

		History:
			06/15/2004	AJR - Moved WAP tags and history here for on-demand support
			06/15/2004	AJR - Moved to a pure LassoScript format
			03/08/2004	FMS - Updated [WAP_IsEnabled] to search only for Accept type.
			06/13/2002	FMS - Converted to LassoScript, Changed to [_Include_URL], Rewrote [Euro]
			02/22/2002	GLH - Fix for difference between IIS and Apache HTTP_ WAP headers
			02/26/2001	GLH - Added WAP tags
			11/22/2000	GLH - Initial work
	*/

	/*
		WAP_IsEnabled:
			Examines MIME headers to see if a WAP device (browser) is hitting the server

		Problems:
			Only search for text/vnd.wap.wml in the Accept header
		Inputs:

		Output:
			true if the browser is a WAP device, false if not
	*/

	Define_Tag: 'IsEnabled', -namespace='_global_WAP_', -Priority='Replace';
		// We return true if the gateway says it accepts 'text/vnd.wap.wml'
		Local: 'temp' = (String_findRegExp: (Client_Headers), -Find='accept\\s*:.*text/vnd.wap.wml', -ignorecase);
		Return: (#temp->Size > 0);
	/Define_Tag;



	/*
		WAP_MaxButtons:
			Examines MIME headers to see how many soft keys the WAP device has

		Problems:
			Could be slow considering _Chop_MIMES is all LDML, and loop search is interpreted
		Inputs:

		Output:
			integer indicating number of softkeys, or undefined if not a WAP device
	*/

	Define_Tag: 'MaxButtons', -namespace='_global_WAP_', -Priority='Replace';
		// First do a quick scan for WML in all the headers, just to rule it out if it doesn't exist
		If: (Client_Headers)->contains: 'UP.B';
			// now we know there's a WML in there somewhere; break into individual MIME headers to see where exactly
			// Chop apart current MIME headers so we can identify HTTP_ACCEPTS
			Local: 'newMIMEs'=(_Chop_MIMES: (Client_Headers));
			If: (Lasso_Version: -LassoPlatform)->contains: 'Win';
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'HTTP_x_up_devcap_numsoftkeys';
						Local: 'NumButtons'=(#newMIMEs->get: Loop_Count)->second;
						Return: #NumButtons;
					/If;
				/Loop;
			Else;
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'x-up-devcap-numsoftkeys';
						Local: 'NumButtons'=(#newMIMEs->get: Loop_Count)->second;
						Return: #NumButtons;
					/If;
				/Loop;
			/If;
		/If;
		Return: 0;
	/Define_Tag;



	/*
		WAP_MaxHorzPixels:
			Examines MIME headers to see how many pixels horizontally the WAP device has

		Problems:
			Could be slow considering _Chop_MIMES is all LDML, and loop search is interpreted
		Inputs:

		Output:
			integer indicating number of horizontal pixels, or undefined if not a WAP device
	*/

	Define_Tag: 'MaxHorzPixels', -namespace='_global_WAP_', -Priority='Replace';
		// First do a quick scan for WML in all the headers, just to rule it out if it doesn't exist
		If: (Client_Headers)->contains: 'UP.B';
			// now we know there's a WML in there somewhere; break into individual MIME headers to see where exactly
			// Chop apart current MIME headers so we can identify HTTP_ACCEPTS
			Local: 'newMIMEs'=(_Chop_MIMES: (Client_Headers));
			If: (Lasso_Version: -LassoPlatform)->contains: 'Win';
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'HTTP_x_up_devcap_screenpixels';
						Local: 'PixelPair'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #PixelPair);
					/If;
				/Loop;
			Else;
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'x-up-devcap-screenpixels';
						Local: 'PixelPair'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #PixelPair);
					/If;
				/Loop;
			/If;
		/If;
		Return: 0;
	/Define_Tag;



	/*
		WAP_MaxVertPixels:
			Examines MIME headers to see how many pixels vertically the WAP device has

		Problems:
			Could be slow considering _Chop_MIMES is all LDML, and loop search is interpreted
		Inputs:

		Output:
			integer indicating number of vertical pixels, or undefined if not a WAP device
	*/

	Define_Tag: 'MaxVertPixels', -namespace='_global_WAP_', -Priority='Replace';
		// First do a quick scan for WML in all the headers, just to rule it out if it doesn't exist
		If: (Client_Headers)->contains: 'UP.B';
			// now we know there's a WML in there somewhere; break into individual MIME headers to see where exactly
			// Chop apart current MIME headers so we can identify HTTP_ACCEPTS
			Local: 'newMIMEs'=(_Chop_MIMES: (Client_Headers));
			If: (Lasso_Version: -LassoPlatform)->contains: 'Win';
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'HTTP_x_up_devcap_screenpixels';
						Local: 'PixelPair'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #PixelPair->substring: (#PixelPair->find: ',') + 1);
					/If;
				/Loop;
			Else;
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'x-up-devcap-screenpixels';
						Local: 'PixelPair'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #PixelPair->substring: (#PixelPair->find: ',') + 1);
					/If;
				/Loop;
			/If;
		/If;
		Return: 0;
	/Define_Tag;



	/*
		WAP_MaxColumns:
			Examines MIME headers to see how many character columns horizontally the WAP device has

		Problems:
			Could be slow considering _Chop_MIMES is all LDML, and loop search is interpreted
		Inputs:

		Output:
			integer indicating number of horizontal character columns, or undefined if not a WAP device
	*/

	Define_Tag: 'MaxColumns', -namespace='_global_WAP_', -Priority='Replace';
		// First do a quick scan for WML in all the headers, just to rule it out if it doesn't exist
		If: (Client_Headers)->contains: 'UP.B';
			// now we know there's a WML in there somewhere; break into individual MIME headers to see where exactly
			// Chop apart current MIME headers so we can identify HTTP_ACCEPTS
			Local: 'newMIMEs'=(_Chop_MIMES: (Client_Headers));
			If: (Lasso_Version: -LassoPlatform)->contains: 'Win';
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'HTTP_x_up_devcap_msize';
						Local: 'msize'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #msize->substring: (#msize->find: ',') + 1);
					/If;
				/Loop;
			Else;
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'x-up-devcap-msize';
						Local: 'msize'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #msize->substring: (#msize->find: ',') + 1);
					/If;
				/Loop;
			/If;
		/If;
		Return: 0;
	/Define_Tag;



	/*
		WAP_MaxRows:
			Examines MIME headers to see how many character columns horizontally the WAP device has

		Problems:
			Could be slow considering _Chop_MIMES is all LDML, and loop search is interpreted
		Inputs:

		Output:
			integer indicating number of horizontal character columns, or undefined if not a WAP device
	*/

	Define_Tag: 'MaxRows', -namespace='_global_WAP_', -Priority='Replace';
		// First do a quick scan for WML in all the headers, just to rule it out if it doesn't exist
		If: (Client_Headers)->contains: 'UP.B';
			// now we know there's a WML in there somewhere; break into individual MIME headers to see where exactly
			// Chop apart current MIME headers so we can identify HTTP_ACCEPTS
			Local: 'newMIMEs'=(_Chop_MIMES: (Client_Headers));
			If: (Lasso_Version: -LassoPlatform)->contains: 'Win';
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'HTTP_x_up_devcap_msize';
						Local: 'msize'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #msize);
					/If;
				/Loop;
			Else;
				Loop: #newMIMEs->size;
					If: (#newMIMEs->get: Loop_Count)->first == 'x-up-devcap-msize';
						Local: 'msize'=(#newMIMEs->get: Loop_Count)->second;
						Return: (integer: #msize);
					/If;
				/Loop;
			/If;
		/If;
		Return: 0;
	/Define_Tag;

?>