<?LassoScript

	// [iCal_Parse] - Parses an arbitrary string and returns an iCal type based on the contents.
	//
	// [iCal_Calendar: "prop"="value", (ical_event), ...] - A calendar with attributes, events, todos, journals, freebusies, and timezones.
	//       ->Events
	//       ->Todos
	//       ->Journals
	//       ->FreeBusies
	//       ->Timezones
	//       ->Attributes
	//       ->Insert
	//   [iCal_Event: "prop"="value", (ical_alarm), ...] - An event with attributes and alarms.
	//       ->Attributes
	//       ->Alarms
	//       ->Standards
	//       ->Daylights
	//       ->Insert
	//     [iCal_Alarm] - An alarm with attributes.
	//   [iCal_ToDo] - A todo with attributes and alarms.
	//     [iCal_Alarm] - An alarm with attributes.
	//   [iCal_Journal] - A journal entry with attributes.
	//   [iCal_FreeBusy] - A freebusy entry with attributes.
	//   [iCal_TimeZone] - A timezone entry with attributes, standard, and daylight.
	//     [iCal_Standard] - A standard timezone setting with attributes.
	//     [iCal_Daylight] - A daylight saving time timezone setting with attributes.
	// 
	// [iCal_Attribute: -Name='', -Value='', -Type='Date/Time', -Params=(array)] - An attribute of any calendar or item with a name, value, and parameters
	//       ->Name
	//       ->Value
	//       ->Params
	//
	// [iCal_Item] - A generic item, parent type of other items.
	//
	
	Define_Tag: 'Parse', -Namespace='iCal_',
			-Required='data';
		if: #data >> 'BEGIN:VCALENDAR';
			return: (ical_calendar: -data=@#data);
		else: #data >> 'BEGIN:VEVENT';
			return: (ical_event: -data=@#data);
		else: #data >> 'BEGIN:VTODO';
			return: (ical_todo: -data=@#data);
		else: #data >> 'BEGIN:VJOURNAL';
			return: (ical_journal: -data=@#data);
		else: #data >> 'BEGIN:VFREEBUSY';
			return: (ical_freebusy: -data=@#data);
		else: #data >> 'BEGIN:VTIMEZONE';
			return: (ical_timezone: -data=@#data);
		else: #data >> 'BEGIN:VALARM';
			return: (ical_alarm: -data=@#data);
		else: #data >> 'BEGIN:STANDARD';
			return: (ical_standard: -data=@#data);
		else: #data >> 'BEGIN:DAYLIGHT';
			return: (ical_daylight: -data=@#data);
		/if;
		fail: -1, 'Could not parse iCal object';
	/Define_Tag;
	
	Define_Type: 'Attribute', -Namespace='iCal_', -Prototype;
		
		Local: 'data';
		Local: 'name' = '';
		Local: 'value' = '';
		Local: 'param' = array;
		local: 'type' = 'raw';
		local: 'format' = '%Y%m%dT%H%M%S';
		Local: 'lazy' = true;
		Local: 'eol' = '\r\n';
		
		Define_Tag: 'onCreate', -optional='name', -optional='value', -optional='type', -optional='params', -optional='data';
			if: (local_defined: 'data');
				Self->'eol' = ((#data >> '\r\n') ? '\r\n' | ((#data >> '\r') ? '\r' | '\n'));
				Self->'data' = #data->(removetrailing: self->'eol') &;
				self->'lazy' = true;
			else;
				self->'name' = (string_uppercase: (local: 'name'));
				self->'value' = (local: 'value');
				if: (local_defined: 'params') && (#params->(isa: 'array'));
					self->'param'->(insertfrom: #params->iterator);
				/if;
				if: ((local: 'value')->(isa: 'pair') && (local: 'value')->first->(isa: 'date'));
					local: 'gmt' = (local: 'value')->first->gmt;
					self->'format' = '%Y%m%dT%H%M%S' + (#gmt  ? 'Z' | '');
				else: (local: 'value')->(isa: 'bytes');
					self->'param'->(removeall: 'value');
					self->'param'->(insert: 'value'='BINARY');
				else: (local: 'value')->(isa: 'duration') && (local: 'name') != 'duration';
					self->'param'->(removeall: 'value');
					self->'param'->(insert: 'value'='DURATION');
				else: (local: 'value')->(isa: 'date');
					local: 'gmt' = (local: 'value')->gmt;
					if: (local: 'type') == 'timestamp';
						self->'format' = '%Y%m%dT%H%M%S' + (#gmt ? 'Z' | '');
						self->'param'->(removeall: 'value');
						self->'param'->(insert: 'value'='TIMESTAMP');
					else: (local: 'type') == 'date';
						self->'format' = '%Y%m%d' + (#gmt  ? 'Z' | '');
						self->'param'->(removeall: 'value');
						self->'param'->(insert: 'value'='DATE');
					else: (local: 'type') == 'time';
						self->'format' = '%H%M%S' + (#gmt  ? 'Z' | '');
						self->'param'->(removeall: 'value');
						self->'param'->(insert: 'value'='TIME');
					else;
						self->'format' = '%Y%m%dT%H%M%S' + (#gmt  ? 'Z' | '');
					/if;
				else: (local: 'type') != '';
					self->'param'->(removeall: 'value');
					self->'param'->(insert: 'value'=#type);
				/if;
				self->'lazy' = false;
			/if;
		/Define_Tag;
		
		Define_Tag: 'name';
			Self->'lazy' ? Self->parse;
			return: @Self->'name';
		/Define_Tag;

		Define_Tag: 'value', -optional='type';
			Self->'lazy' ? Self->parse;
			return: @Self->'value';
		/Define_Tag;
		
		Define_Tag: 'params';
			Self->'lazy' ? Self->parse;
			return: @Self->'param';
		/Define_Tag;
		
		Define_Tag: 'valuetype';
			Self->'lazy' ? Self->parse;
			return: @Self->'type';
		/Define_Tag;
		
		Define_Tag: 'parse';
			Local: 'parts' = (String_FindRegExp: Self->'data', -Find='^([A-z0-9-]+)((?:;[A-z0-9-]+=(?:[^";:,]+|"[^"]+"))*):(.+)$');
			Fail_If(!#parts->Size, -1, "Malformed iCal attribute: " + Self->'data');
			Self->'name' = #parts->(Get: 2);
			Self->'data' = #parts->(Get: 4);
			Local: 'paramsRegExp' = (RegExp: -Input=#parts->(Get: 3), -Find=';([A-z0-9-]+)=(?:([^";:,]+)|"([^"]+)")');
			While: #paramsRegExp->Find;
				Self->'param'->(Insert: #paramsRegExp->(MatchString: 1) = (#paramsRegExp->(MatchString: 2) || #paramsRegExp->(MatchString: 3)));
			/While;			
/* 
			Self->'name' = (string_findregexp: Self->'data', -ignorecase, -find='[^;=:,"]+[;:]')->first;
			Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find='[^;=:,"]+[;:]', -replace='', -replaceonlyone);
			if: Self->'name'->(endswith: ';');
				Local: 'array' = (string_findregexp: Self->'data', -ignorecase, -find='([^;=:,]+)="([^"]+)"(?=[;:])');
				loop: #array->size, -by=3; 
					Local: 'pair' = (pair: #array->(get: loop_count + 1) = #array->(get: loop_count + 2));
					Self->'param'->(insert: #pair);
				/loop;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find='[^;=:,]+="[^"]+"(?=[;:])', -replace='');
				Local: 'array' = (string_findregexp: Self->'data', -ignorecase, -find='([^;=:,]+)=([^;=:,]+)(?=[;:])');
				loop: #array->size, -by=3; 
					Local: 'pair' = (pair: #array->(get: loop_count + 1) = #array->(get: loop_count + 2));
					Self->'param'->(insert: #pair);
				/loop;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find='[^;=:,]+=[^;=:,]+(?=[;:])', -replace='');
			/if;
			Self->'name' -= ';';
			Self->'name' -= ':';
			Self->'data'->(removeleading: ':');
			Self->'data'->(removeleading: ';');
*/
			Self->'data'->(replace: '\\N', Self->'eol');
			Self->'data'->(replace: '\\\\', '\\');
			Self->'data'->(replace: '\\;', ';');
			Self->'data'->(replace: '\\,', ',');
			local: 'temp' = self->'param'->(find: 'VALUE');
			if: #temp->size > 0;
				self->'type' = #temp->first->second;
			else: (self->'name' == 'dtstamp');
				self->'type' = 'TIMESTAMP';
			else: (self->'name' == 'dtstart') || (self->'name' == 'dtend');
				self->'type' = 'DATE-TIME';
			else: (self->'name' == 'dtstamp');
				self->'type' = 'TIMESTAMP';
			else: (self->'name' == 'rrule');
				self->'type' = 'RECUR';
			else: (self->'name' == 'duration') || (self->'name' == 'trigger');
				self->'type' = 'DURATION';
			else;
				self->'type' = 'RAW';
			/if;
			if: self->'type' == 'BINARY';
				self->'value' = (Decode_Base64: Self->'data');
			else: self->'type' == 'BOOLEAN';
				self->'value' = (Boolean: Self->'data');
			else: self->'type' == 'CAL-ADDRESS';
				self->'value' = (String: Self->'data');
			else: self->'type' == 'DATE';
				self->'value' = (Date: Self->'data', -Format=(self->'format' := '%Y%m%d'));
			else: self->'type' == 'DATE-TIME';
				if: self->'data' >> 'Z';
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%Y%m%dT%H%M%SZ'), -gmt);
				else;
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%Y%m%dT%H%M%S'));
				/if;
			else: self->'type' == 'DURATION';
				local: 'hour' = 0;
				local: 'minute' = 0;
				local: 'second' = 0;
				Self->'data' >> 'W' ? #hour += 168 * (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)W.*', -replace='\\1'));
				Self->'data' >> 'D' ? #hour += 24 * (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)D.*', -replace='\\1'));
				Self->'data' >> 'H' ? #hour += (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)H.*', -replace='\\1'));
				Self->'data' >> 'M' ? #minute += (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)M.*', -replace='\\1'));
				Self->'data' >> 'S' ? #second = (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)S.*', -replace='\\1'));
				self->'value' = (duration: -hour=#hour, -minute=#minute, -second=#second);
			else: self->'type' == 'FLOAT';
				self->'value' = (Float: Self->'data');
			else: self->'type' == 'INTEGER';
				self->'value' = (Integer: Self->'data');
			else: self->'type' == 'PERIOD';
				local: 'start';
				local: 'split' = self->'data'->(split: '/');
				if: #split->first >> 'Z';
					local: 'start' = (Date: #split->first, -Format=(self->'format' := '%Y%m%dT%H%M%SZ'), -gmt);
				else;
					local: 'start' = (Date: #split->first, -Format=(self->'format' := '%Y%m%dT%H%M%S'));
				/if;
				if: #split->second >> 'P';
					local: 'hour' = 0;
					local: 'minute' = 0;
					local: 'second' = 0;
					local: 'factor' = ((self->'data' >> '-P') ? -1 | 1);
					Self->'data' >> 'W' ? #hour += 168 * (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)W.*', -replace='\\1'));
					Self->'data' >> 'D' ? #hour += 24 * (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)D.*', -replace='\\1'));
					Self->'data' >> 'H' ? #hour += (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)H.*', -replace='\\1'));
					Self->'data' >> 'M' ? #minute += (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)M.*', -replace='\\1'));
					Self->'data' >> 'S' ? #second = (integer: (string_replaceregexp: Self->'data', -find='P.*?(\\d+)S.*', -replace='\\1'));
					self->'value' = (pair: #start = (duration: -hour=(#factor * #hour), -minute=(#factor * #minute), -second=(#factor * #second)));
				else;
					if: #split->first >> 'Z';
						local: 'end' = (Date: #split->first, -Format=(self->'format' := '%Y%m%dT%H%M%SZ'), -gmt);
					else;
						local: 'end' = (Date: #split->first, -Format=(self->'format' := '%Y%m%dT%H%M%S'));
					/if;
					self->'value' = (pair: #start = #end);
				/if;
			else: self->'type' == 'RECUR';
				local: 'recur' = (map:
					'INTERVAL' = 1);
				local: 'prior' = '';
				iterate: self->'data'->(split: ';'), (local: 'param');
					if: #param >> '=';
						local: 'name' = #param->(split: '=')->first;
						local: 'value' = #param->(split: '=')->second;
						if: #name == 'freq';
							if: (array: 'yearly','monthly','daily','hourly','minutely','secondly') >> #value;
								#recur->(insert: 'FREQ' = #value);
							/if;
						else: #name == 'count';
							#recur->(insert: 'COUNT' = (integer: #value));
						else: #name == 'interval';
							#recur->(insert: 'INTERVAL' = (integer: #value));
						else: #name == 'until';
							if: #value >> 'Z';
								#recur->(insert: 'UNTIL' = (Date: #value, -Format=(self->'format' := '%Y%m%dT%H%M%SZ'), -gmt));
							else;
								#recur->(insert: 'UNTIL' = (Date: #value, -Format=(self->'format' := '%Y%m%d')));
							/if;
						else: #name >> 'bysetpos';
							#recur->(insert: (#prior + 'POS') = #value->(split:','));
						else: #name == 'wkst' || #name >> 'by';
							#recur->(insert: (string_uppercase: #name) = #value->(split:','));
						/if;
						local: 'prior' = #name;
					/if;
				/iterate;
				self->'value' = #recur;
			else: self->'type' == 'TEXT';
				self->'value' = (String: Self->'data');
			else: self->'type' == 'TIME';
				if: self->'data' >> 'Z';
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%H%M%SZ'), -gmt);
				else;
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%H%M%S'));
				/if;
			else: self->'type' == 'TIMESTAMP';
				if: self->'data' >> 'Z';
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%Y%m%dT%H%M%SZ'), -gmt);
				else;
					self->'value' = (Date: Self->'data', -Format=(self->'format' := '%Y%m%dT%H%M%S'));
				/if;
			else: self->'type' == 'URI';
				self->'value' = (String: Self->'data');
			else: self->'type' == 'UTC-OFFSET';
				self->'value' = (String: Self->'data');
			else;
				self->'value' = (String: Self->'data');
			/if;
			Self->'data' = '';
			Self->'lazy' = false;
		/Define_Tag;
		
		Define_Tag: 'onConvert';
			if: Self->'lazy';
				return: Self->'data' + Self->'eol';
			else;
				Local: 'output' = '';
				#output += Self->'name';
				iterate: Self->'param', (Local: 'temp');
					if: (#temp->second >> ':') || (#temp->second >> ';') || (#temp->second >> '"');
						#output += ';' + (string_uppercase: #temp->first) + '="' + #temp->second + '"';
					else;
						#output += ';' + (string_uppercase: #temp->first) + '=' + #temp->second;
					/if;
				/iterate;
				#output += ':';
				if: self->'value'->(isa: 'pair');
					if: self->'value'->first->(isa: 'date');
						#output += Self->'value'->first->(format: self->'format') + '/';
					else;
						#output += (string: Self->'value') + '/';
					/if;
					if: self->'value'->second->(isa: 'date');
						#output += Self->'value'->second->(format: self->'format');
					/if;
				/if;
				if: self->'value'->(isa: 'string');
					local: 'temp' = self->'value';
					#temp->(replace: '\\', '\\\\');
					#temp->(replace: ';', '\\;');
					#temp->(replace: ',', '\\,');
					#temp = (string_replaceregexp: #temp, -find='\r\n|\r|\n', -replace='\\N');
					#output += #temp;
				else: self->'value'->(isa: 'boolean');
					#output += (Self->'value' ? 'TRUE' | 'FALSE');
				else: self->'value'->(isa: 'bytes');
					#output += (Encode_Base64: Self->'value');
				else: self->'value'->(isa: 'date');
					#output += Self->'value'->(format: self->'format');
				else: self->'value'->(isa: 'decimal');
					#output += Self->'value';
				else: self->'value'->(isa: 'duration') || ( (self->'value'->(isa: 'pair') && self->'value'->second->(isa: 'duration')) );
					local: 'hour' = self->'value'->(isa: 'duration') ? self->'value'->second | self->'value'->second->second;
					#output += (#hour < 0 ? '-' | '') + 'P';
					#hour < 0 ? #hour *= -1;
					local: 'second' = #hour % 60;
					#hour = (#hour - #second) / 60;
					local: 'minute' = #hour % 60;
					#hour = (#hour - #minute) / 60;
					local: 'day' = #hour / 24;
					#hour = #hour - (#day * 24);
					if: (#day % 7 == 0) && (#hour == 0) && (#minute == 0) && (#second == 0);
						#output += (#day / 7) + 'W';
					else: (#day > 0) || (#hour >> 0) || (#minute >> 0) || (#second >> 0);
						(#day > 0) ? #output += #day + 'D';
						if: (#hour > 0) || (#minute > 0) || (#second > 0);
							#output += 'T';
							(#hour > 0) ?  #output += #hour + 'H';
							(#minute > 0) ? #output += #minute + 'M';
							(#second > 0) ? #output += #second + 'S';
						/if;
					/if;
				else: self->'value'->(isa: 'integer');
					#output += Self->'value';
				else: self->'value'->(isa: 'map');
					#output += 'FREQ=' + (string_uppercase: self->'value'->(find: 'freq'));
					if: self->'value'->(find: 'interval') > 0;
						#output += ';INTERVAL=' + self->'value'->(find: 'interval');
					/if;
					if: self->'value'->(find: 'count') > 0;
						#output += ';COUNT=' + self->'value'->(find: 'count');
					/if;
					if: self->'value'->(find: 'until') != '';
						local: 'date' = (date: self->'value'->(find: 'until'));
						if: #date->gmt;	
							#output += ';UNTIL=' + #date->(format: '%Y%m%dT%H%M%SZ');
						else;
							#output += ';UNTIL=' + #date->(format: '%Y%m%d');
						/if;
					/if;
					iterate: self->'value'->keys, (local: 'name');
						if: #name->(beginswith: 'by') || #name == 'wkst';
							(#name >> 'pos') ? loop_continue;
							local: 'value' = self->'value'->(find: #name);
							if: #value->(isa: 'array') && (#value->size > 0);
								#output += ';' + (string_uppercase: #name) + '=' + #value->(join: ',');
								if: self->'value'->(find: #name + 'pos');
									local: 'value' = self->'value'->(find: #name + 'pos');
									if: #value->(isa: 'array') && (#value->size > 0);
										#output += ';BYSETPOS=' + #value->(join: ',');
									/if;
								/if;
							else: #value != '';
								#output += ';' + (string_uppercase: #name) + '=' + #value;
							/if;
						/if;
					/iterate;
				else: !(self->'value'->(isa: 'pair'));
					#output += (String: Self->'value');
				/if;
				If: #output->size > 75;
					#output = (string_replaceregexp: #output, -find='(.{75})', -replace='\\1' + self->'eol' + ' ');
				/If;
				#output += Self->'eol';
				return: @#output;
			/if;
		/Define_Tag;
		
		Define_Tag: 'onCompare', -required='value';
			return: (string: self) == (string: #value) ? 0 | 1;
		/Define_Tag;
		
	/Define_Type;	
	
	Define_Type: 'Item', -Namespace='iCal_', -Prototype;
		
		Local: 'key' = 'VEVENT';
		Local: 'data';
		Local: 'attribute' = array;
		Local: 'standard' = array;
		Local: 'daylight' = array;
		Local: 'alarm' = array;
		Local: 'lazy' = true;
		Local: 'eol' = '\r\n';

		Define_Tag: 'onCreate';
			if: (local_defined: 'data');
				Self->'eol' = ((#data >> '\r\n') ? '\r\n' | ((#data >> '\r') ? '\r' | '\n'));
				Self->'data' = #data;
			else;
				self->(insert: params);
				local: 'array' = array;
				iterate: self->'attribute', (local: 'temp');
					#array->(insert: #temp->name);
				/iterate;
				Self->'data' = '';
			/if;
		/Define_Tag;

		Define_Tag: 'attributes';
			Self->'data'->Find('BEGIN:VALARM') > 1 ? Self->alarms;
			Self->'data'->Find('BEGIN:STANDARD') > 1 ? Self->standards;
			Self->'data'->Find('BEGIN:DAYLIGHT') > 1 ? Self->daylights;
			if: (Self->'data' != '');
				Local: 'pattern' = '(?s).+?(?-s)[\r\n]+';
				Self->'data' = (string_replaceregexp: Self->'data', -find='[\r\n]+[ \t]', -replace='');
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					if: !(#temp->(beginswith: 'BEGIN:') || #temp->(beginswith: 'END:'));
						Self->'attribute'->(insert: (ical_attribute: -data=#temp));
					/if;
				/iterate;
				Self->'data' = '';
				Self->'lazy' = false;
			/if;
			local: 'uid' = false;
			iterate: self->'attribute', (local: 'temp');
				#uid = #uid || (#temp->name == 'uid');
			/iterate;
			#uid == false ? self->'attribute'->(insert: (ical_attribute: 'uid', lasso_uniqueid));
			return: @Self->'attribute';
		/Define_Tag;
		
		Define_Tag: 'getattribute', -required='name';
			local: 'attributes' = self->attributes;
			local: 'output' = array;
			loop: #attributes->size;
				#attributes->(get: loop_count)->name == #name ? #output->(insert: @(#attributes->(get: loop_count)));
			/loop;
			if: #output->size > 1;
				return: @#output;
			else;
				return: @(#output->first);
			/if;
		/Define_Tag;
		
		Define_Tag: 'hasattribute', -required='name';
			local: 'attributes' = self->attributes;
			loop: #attributes->size;
				#attributes->(get: loop_count)->name == #name ? return: true;
			/loop;
			return: false;
		/Define_Tag;
		
		Define_Tag: 'addattribute', -optional='name', -optional='value', -optional='type', -optional='params';
			self->'attribute'->(insert: (ical_attribute: -name=(local: 'name'), -value=(local: 'value'), -type=(local: 'type'), -params=(local: 'params')));
		/Define_Tag;
		
		Define_Tag: 'removeattribute', -required='name';
			local: 'attributes' = self->attributes;
			loop: -from=#attributes->size, -to=1, -by=-1;
				#attributes->(get: loop_count)->name == #name ? #attributes->(remove: loop_count);
			/loop;
		/Define_Tag;
		
		Define_Tag: 'alarms';
			if: (Self->'data' >> 'BEGIN:VALARM');
				Local: 'pattern' = 'BEGIN:VALARM\\s+(?s).+?(?-s)\\s+END:VALARM\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'alarm'->(insert: (ical_alarm: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'alarm';
		/Define_Tag;

		Define_Tag: 'standards';
			if: (Self->'data' >> 'BEGIN:STANDARD');
				Local: 'pattern' = 'BEGIN:STANDARD\\s+(?s).+?(?-s)\\s+END:STANDARD\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'standard'->(insert: (ical_timezone: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'standard';
		/Define_Tag;

		Define_Tag: 'daylights';
			if: (Self->'data' >> 'BEGIN:DAYLIGHT');
				Local: 'pattern' = 'BEGIN:DAYLIGHT\\s+(?s).+?(?-s)\\s+END:DAYLIGHT\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'daylight'->(insert: (ical_timezone: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'daylight';
		/Define_Tag;
		
		Define_Tag: 'Insert', -required='data';
			if: #data->(isa: 'string');
				Self->'data' += #data;
			else: #data->(isa: 'array');
				iterate: #data, (Local: 'temp');
					Self->(insert: #temp);
				/iterate;
			else: #data->(isa: 'ical_attribute');
				Self->'attribute'->(insert: #data);
			else: #data->(isa: 'ical_alarm');
				Self->'alarm'->(insert: #data);
			else: #data->(isa: 'ical_standard');
				Self->'standard'->(insert: #data);
			else: #data->(isa: 'ical_daylight');
				Self->'daylight'->(insert: #data);
			else: #data->(isa: 'pair');
				self->'attribute'->(insert: (ical_attribute: -name=#data->first, -value=#data->second));
			/if;
			Self->'lazy' = false;
		/Define_Tag;

		Define_Tag: 'onConvert';
			if: Self->'lazy';
				return: Self->'data';
			else;
				Self->attributes;
				Local: 'output' = 'BEGIN:' + Self->'key' + Self->'eol';
				#output += Self->'attribute'->(join: '');
				#output += Self->'alarm'->(join: '');
				#output += Self->'standard'->(join: '');
				#output += Self->'daylight'->(join: '');
				#output += 'END:' + Self->'key' + Self->'eol';
				return: @#output;
			/if;
		/Define_Tag;

		Define_Tag: 'onCompare', -required='value';
			return: (string: self) == (string: #value) ? 0 | 1;
		/Define_Tag;

	/Define_Type;

	Define_Type: 'Event', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VEVENT';
	/Define_Type;	

	Define_Type: 'ToDo', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VTODO';
	/Define_Type;

	Define_Type: 'Journal', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VJOURNAL';
	/Define_Type;

	Define_Type: 'FreeBusy', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VFREEBUSY';
	/Define_Type;

	Define_Type: 'TimeZone', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VTIMEZONE';
	/Define_Type;

	Define_Type: 'Alarm', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'VALARM';
	/Define_Type;

	Define_Type: 'Standard', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'STANDARD';
	/Define_Type;

	Define_Type: 'Daylight', -Namespace='iCal_', -Prototype, 'iCal_Item';
		Local: 'key' = 'Daylight';
	/Define_Type;

	Define_Type: 'Calendar', -Namespace='iCal_', -Prototype;
		
		Local: 'data';
		Local: 'attribute' = array;
		Local: 'event' = array;
		Local: 'todo' = array;
		Local: 'journal' = array;
		Local: 'freebusy' = array;
		Local: 'timezone' = array;
		Local: 'lazy' = true;
		Local: 'eol' = '\r\n';
		
		Define_Tag: 'onCreate';
			if: (local_defined: 'data');
				Self->'eol' = ((#data >> '\r\n') ? '\r\n' | ((#data >> '\r') ? '\r' | '\n'));
				Self->'data' = #data;
			else;
				self->(insert: params);
				local: 'array' = array;
				iterate: self->'attribute', (local: 'temp');
					#array->(insert: #temp->name);
				/iterate;
				#array !>> 'version' ? self->'attribute'->(insert: (ical_attribute: -name='version', -value='2.0'));
				#array !>> 'prodid' ? self->'attribute'->(insert: (ical_attribute: -name='prodid', -value='Lasso ' (lasso_version: -lassoversion)));
			/if;
		/Define_Tag;
	
		Define_Tag: 'events';
			if: (Self->'data' >> 'BEGIN:VEVENT');
				Local: 'pattern' = 'BEGIN:VEVENT\\s+(?s).+?(?-s)\\s+END:VEVENT\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'event'->(insert: (ical_event: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'event';
		/Define_Tag;

		Define_Tag: 'todos';
			if: (Self->'data' >> 'BEGIN:TODO');
				Local: 'pattern' = 'BEGIN:VTODO\\s+(?s).+?(?-s)\\s+END:VTODO\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'todo'->(insert: (ical_todo: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'todo';
		/Define_Tag;
	
		Define_Tag: 'journals';
			if: (Self->'data' >> 'BEGIN:VJOURNAL');
				Local: 'pattern' = 'BEGIN:VJOURNAL\\s+(?s).+?(?-s)\\s+END:VJOURNAL\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'journal'->(insert: (ical_journal: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'journal';
		/Define_Tag;

		Define_Tag: 'freebusies';
			if: (Self->'data' >> 'BEGIN:VFREEBUSY');
				Local: 'pattern' = 'BEGIN:VFREEBUSY\\s+(?s).+?(?-s)\\s+END:VFREEBUSY\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'freebusy'->(insert: (ical_freebusy: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'freebusy';
		/Define_Tag;

		Define_Tag: 'timezones';
			if: (Self->'data' >> 'BEGIN:VTIMEZONE');
				Local: 'pattern' = 'BEGIN:VTIMEZONE\\s+(?s).+?(?-s)\\s+END:VTIMEZONE\\s*';
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					Self->'timezone'->(insert: (ical_timezone: -data=#temp));
				/iterate;
				Self->'data' = (string_replaceregexp: Self->'data', -ignorecase, -find=#pattern, -replace='');
				Self->'lazy' = false;
			/if;
			return: @Self->'timezone';
		/Define_Tag;

		Define_Tag: 'attributes';
			(Self->'data' >> 'BEGIN:VEVENT') ? Self->events;
			(Self->'data' >> 'BEGIN:VTODO') ? Self->todos;
			(Self->'data' >> 'BEGIN:VJOURNAL') ? Self->journals;
			(Self->'data' >> 'BEGIN:VFREEBUSY') ? Self->freebusys;
			(Self->'data' >> 'BEGIN:VTIMEZONE') ? Self->timezones;
			if: (Self->'data' != '');
				Local: 'pattern' = '(?s).+?(?-s)[\r\n]+';
				Self->'data' = (string_replaceregexp: Self->'data', -find='[\r\n]+[ \t]', -replace='');
				iterate: (string_findregexp: Self->'data', -ignorecase, -find=#pattern), (Local: 'temp');
					if: !(#temp->(beginswith: 'BEGIN:') || #temp->(beginswith: 'END:'));
						Self->'attribute'->(insert: (ical_attribute: -data=#temp));
					/if;
				/iterate;
				Self->'data' = '';
				Self->'lazy' = false;
			/if;
			return: @Self->'attribute';
		/Define_Tag;
			
		Define_Tag: 'getattribute', -required='name';
			local: 'attributes' = self->attributes;
			local: 'output' = array;
			loop: #attributes->size;
				#attributes->(get: loop_count)->name == #name ? #output->(insert: @(#attributes->(get: loop_count)));
			/loop;
			if: #output->size > 1;
				return: @#output;
			else;
				return: @(#output->first);
			/if;
		/Define_Tag;
		
		Define_Tag: 'hasattribute', -required='name';
			local: 'attributes' = self->attributes;
			loop: #attributes->size;
				#attributes->(get: loop_count)->name == #name ? return: true;
			/loop;
			return: false;
		/Define_Tag;
		
		Define_Tag: 'addattribute', -optional='name', -optional='value', -optional='type', -optional='params';
			self->'attribute'->(insert: (ical_attribute: -name=(local: 'name'), -value=(local: 'value'), -type=(local: 'type'), -params=(local: 'params')));
		/Define_Tag;
		
		Define_Tag: 'removeattribute', -required='name';
			local: 'attributes' = self->attributes;
			loop: -from=#attributes->size, -to=1, -by=-1;
				#attributes->(get: loop_count)->name == #name ? #attributes->(remove: loop_count);
			/loop;
		/Define_Tag;

		Define_Tag: 'Insert', -required='data';
			if: #data->(isa: 'string');
				Self->'data' += #data;
			else: #data->(isa: 'array');
				iterate: #data, (Local: 'temp');
					Self->(insert: #temp);
				/iterate;
			else: #data->(isa: 'ical_event');
				Self->'event'->(insert: #data);
			else: #data->(isa: 'ical_todo');
				Self->'todo'->(insert: #data);
			else: #data->(isa: 'ical_journal');
				Self->'journal'->(insert: #data);
			else: #data->(isa: 'ical_freebusy');
				Self->'freebusy'->(insert: #data);
			else: #data->(isa: 'ical_timezone');
				Self->'timezone'->(insert: #data);
			else: #data->(isa: 'ical_attribute');
				Self->'attribute'->(insert: #data);
			else: #data->(isa: 'pair');
				self->'attribute'->(insert: (ical_attribute: -name=#data->first, -value=#data->second));
			/if;
			Self->'lazy' = false;
		/Define_Tag;

		Define_Tag: 'onConvert';
			if: Self->'lazy';
				return: Self->'data';
			else;
				Self->attributes;
				Local: 'output' = 'BEGIN:VCALENDAR' + Self->'eol';
				#output += Self->'attribute'->(join: '');
				#output += Self->'timezone'->(join: '');
				#output += Self->'event'->(join: '');
				#output += Self->'todo'->(join: '');
				#output += Self->'journal'->(join: '');
				#output += Self->'freebusy'->(join: '');
				#output += 'END:VCALENDAR' + Self->'eol';
				return: @#output;
			/if;
		/Define_Tag;
		
		Define_Tag: 'onCompare', -required='value';
			return: (string: self) == (string: #value) ? 0 | 1;
		/Define_Tag;

	/Define_Type;
	
?>
